using System;
using System.Collections;
using System.Xml;

namespace pico.ap
{
  public class ApplicationProfile : IEnumerable
  {
    XmlDocument _xmlDoc = null;
    XmlNamespaceManager _xmlnsMgr = null;
    XmlNode _nodeAbout = null;
    SortedList _termUsageList = null;

    public class TermUsageException : ApplicationException
    {
      public TermUsageException()
      {
      }
      public TermUsageException(string message)
        : base(message)
      {
      }
      public TermUsageException(string message, Exception inner)
        : base(message, inner)
      {
      }
    }

    public class TermUsage
    {
      public enum TermUsageType {Undefined, Element, ElementRefinement, EncodingScheme};
      public enum TermUsageObligation {Undefined, Mandatory, Optional, Recommended};
      public enum TermUsageDataType {Undefined, String}

      string _uri = "";
      string _name = "";
      TermUsageType _type = TermUsageType.Undefined;
      string _definedBy = "";
      TermUsageDataType _dataType = TermUsageDataType.Undefined;
      TermUsageObligation _obligation = TermUsageObligation.Undefined;
      string _refines = null;
      string[] _encodingSchemes = null;
      //string[] _inheritedEncodingSchemes = null;
      
      public TermUsage baseTermUsage = null;

      string[] _encodingSchemeFor = null;
      //string[] _elementRefinements = null;
      
      //SortedList _encodingSchemeList = null;
      //SortedList _elementRefinementList = null;
      
      public TermUsage()
      {
        //_encodingSchemeList = new SortedList();
        //_elementRefinementList = new SortedList();
      }

      public TermUsageType GetTermType()
      {
        return this._type;
      }

      public static TermUsageType StringToTermType(string sType)
      {
        if(sType.Equals("encoding scheme"))
          return TermUsageType.EncodingScheme;

        if(sType.Equals("element"))
          return TermUsageType.Element;

        if(sType.Equals("element refinement"))
          return TermUsageType.ElementRefinement;
        
        throw new TermUsageException("Term Type not valid");
      }

      public static TermUsageDataType GetDataType(string sDataType)
      {
        if(sDataType.Equals("String"))
          return TermUsageDataType.String;

        throw new TermUsageException("Term DataType not valid");
      }

      public static TermUsageObligation GetObligation(string sType)
      {
        if(sType.Equals("Optional"))
          return TermUsageObligation.Optional;

        if(sType.Equals("Recommended"))
          return TermUsageObligation.Recommended;

        if(sType.Equals("Mandatory"))
          return TermUsageObligation.Mandatory;
        
        throw new TermUsageException("Term Obligation not valid");
      }

      public bool HasEncodingScheme(string uri)
      {
        if(this.EncodingSchemes == null)
          return false;

        foreach(string _tmp in this.EncodingSchemes)
        {
          if(_tmp.Equals(uri))
            return true;
        }
        
        if(this.baseTermUsage != null)
        {
          if(this.baseTermUsage.HasEncodingScheme(uri))
            return true;
        }

        return false;
      }

      public string Uri
      {
        get
        { 
          return _uri;
        }

        set
        {
          _uri = value;
        }
      }

      public string Name
      {
        get
        { 
          return _name;
        }

        set
        {
          _name = value;
        }
      }

      public TermUsageType Type
      {
        get
        { 
          return _type;
        }

        set
        {
          _type = value;
        }
      }

      public TermUsageDataType DataType
      {
        get
        { 
          return _dataType;
        }

        set
        {
          _dataType = value;
        }
      }

      public TermUsageObligation Obligation
      {
        get
        { 
          return _obligation;
        }

        set
        {
          _obligation = value;
        }
      }

      public string DefinedBy
      {
        get
        { 
          return _definedBy;
        }

        set
        {
          _definedBy = value;
        }
      }

      public string Refines
      {
        get
        {
          return _refines;
        }

        set
        {
          _refines = value;
        }
      }

      /*
      public string[] ElementRefinements
      {
        get
        {
          return this._elementRefinements;
        }

        set
        {
          this._elementRefinements = value;
        }
      }
      */

      public string[] EncodingSchemes
      {
        get
        {
          return this._encodingSchemes;
        }

        set
        {
          this._encodingSchemes = value;
        }
      }

      public string[] EncodingSchemeFor
      {
        get
        {
          return this._encodingSchemeFor;
        }

        set
        {
          this._encodingSchemeFor = value;
        }
      }

      public void AddEncodingScheme(string termUri)
      {
        ArrayList arr = new ArrayList();
        if(this._encodingSchemes != null)
          arr.AddRange(this._encodingSchemes);
        arr.Add(termUri);

        this._encodingSchemes = (string[])arr.ToArray(typeof(string));
      }      
      
      /*
      public bool HasEncodingScheme(string termName)
      {
        if(this._type == TermUsageType.Element ||
          this._type == TermUsageType.ElementRefinement)
        {
          return this._encodingSchemeList.ContainsKey(termName);
        }
        return false;
      }
      
      public bool IsEncodingSchemeFor(string termName)
      {
        if(this._type == TermUsageType.EncodingScheme)
        {
          return this._encodingSchemeList.ContainsKey(termName);
        }
        return false;
      }
      */
    }

    private string GetNodeText(XmlNode node, string xpath)
    {
      XmlNode nodeTmp = node.SelectSingleNode(xpath, _xmlnsMgr);
      
      string text = "";
      
      if(nodeTmp != null)
        text = nodeTmp.InnerText;
          
      return text;
    }

    private string[] GetNodeTextArray(XmlNode node, string xpath)
    {
      XmlNodeList nodeList = node.SelectNodes(xpath, _xmlnsMgr);

      if(nodeList != null)
      {
        int nodeCount = nodeList.Count;
        if(nodeCount > 0)
        {
          ArrayList stringList = new ArrayList();
          foreach(XmlNode n in nodeList)
          {
            if(n.InnerText.Length > 0)
              stringList.Add(n.InnerText);
          }
          if(stringList.Count > 0)
            return (string[])stringList.ToArray(typeof(string));
        }
      }

      return null;
    }

    public ApplicationProfile(XmlDocument xmlDoc)
    {
      _xmlDoc = xmlDoc;
     
      _termUsageList = new SortedList();

      try
      {
        //Crea un oggetto XmlNamespaceManager per gestire i namespace del PICOAP. 
        _xmlnsMgr = new System.Xml.XmlNamespaceManager(_xmlDoc.NameTable);

        //Aggiunge i namespace del PICOAP
        _xmlnsMgr.AddNamespace("dcap", "http://pico.sns.it/pico/dcap/");
        _xmlnsMgr.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");
        _xmlnsMgr.AddNamespace("dcterms", "http://purl.org/dc/terms/");
        _xmlnsMgr.AddNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance");
        
        //Esegue una query XPath per recuperare il nodo about.        
        _nodeAbout = _xmlDoc.SelectSingleNode("//dcap:about", _xmlnsMgr);
        
        //Dichiara un XmlNodeList per contenere tutti i nodi "term" del PICOAP
        System.Xml.XmlNodeList nodeList;
        
        //Esegue la query XPath per recuperare tutti i nodi "term" dichiarati nel PICOAP
        nodeList = _xmlDoc.SelectNodes("//dcap:term", _xmlnsMgr);
        
        //Itera su tutti i nodi
        foreach (System.Xml.XmlNode node in nodeList)
        {
          string termName = GetNodeText(node, "./dcap:name");
          string termUri = GetNodeText(node, "./dcap:uri");
          string termType = GetNodeText(node, "./dcap:type");
          string termDataType = GetNodeText(node, "./dcap:datatype");
          string termObligation = GetNodeText(node, "./dcap:obligation");
          string termDefinedBy = GetNodeText(node, "./dcap:definedBy");
          string termRefines = GetNodeText(node, "./dcap:refines");
          string[] termHasEncodingScheme = GetNodeTextArray(node, "./dcap:hasEncodingScheme");
          //string[] termIsEncodingSchemeFor = GetNodeTextArray(node, "./dcap:isEncodingSchemeFor");
          string[] termEncodingSchemeFor = GetNodeTextArray(node, "./dcap:encodingSchemeFor");
          string[] termRefinedBy = GetNodeTextArray(node, "./dcap:refinedBy");
                   
          TermUsage termUsage = new TermUsage();
                    
          termUsage.Name = termName;
          termUsage.Uri = termUri;
          termUsage.Type = TermUsage.StringToTermType(termType);
          termUsage.DataType = TermUsage.GetDataType(termDataType);
          termUsage.Obligation = TermUsage.GetObligation(termObligation);
          termUsage.DefinedBy = termDefinedBy;
          termUsage.Refines = termRefines;
          
          // Se  un element pu avere degli encoding scheme
          if(termUsage.Type == TermUsage.TermUsageType.Element ||
            termUsage.Type == TermUsage.TermUsageType.ElementRefinement)
          {
            termUsage.EncodingSchemes = termHasEncodingScheme;
            
          } // Se  un encoding scheme pu avere gli elementi che lo usano
          else if(termUsage.Type == TermUsage.TermUsageType.EncodingScheme)
          {
            termUsage.EncodingSchemeFor = termEncodingSchemeFor;
          }
          
          //this._termUsageList.Add(termName, termUsage);
          this._termUsageList.Add(termUri, termUsage);
        }
				        
        // Un element refinement eredita gli encoding scheme dell'element
        // da cui deriva per cui vengono processati tutti gli element refinement
        // per caricare anche l'element da cui deriva...
        //foreach(string termName in this._termUsageList.Keys)
        foreach(string termUri in this._termUsageList.Keys)
        {
          //TermUsage tu = (TermUsage)this._termUsageList[termName];
          TermUsage tu = (TermUsage)this._termUsageList[termUri];

          if(tu.GetTermType() == TermUsage.TermUsageType.ElementRefinement)
          {
            //if(this._termUsageList.ContainsKey(baseUri))
            TermUsage baseTermUsage = FindTermByUri(tu.Refines);
            if(baseTermUsage != null)
              tu.baseTermUsage = baseTermUsage;            
          }
        }
      }
      catch(Exception exc)
      {
        string msg = exc.Message;
        int i = 0;
        i++;
      }            
    }

    public void AddTermUsage(TermUsage termUsage)
    {
      //string termName = termUsage.Name;
      string _termUri = termUsage.Uri;

      //if(this._termUsageList.ContainsKey(termName))
      if(this._termUsageList.ContainsKey(_termUri))
        throw new TermUsageException("TermUsage gi presente");

      if(termUsage.Type == TermUsage.TermUsageType.EncodingScheme)
      {
        foreach(string termUri in termUsage.EncodingSchemeFor)
        {
          TermUsage termUsageElem = this.FindTermByUri(termUri);
          if(termUsageElem != null)
          {
            if(!termUsageElem.HasEncodingScheme(termUri))
              termUsageElem.AddEncodingScheme(termUsage.Uri);
          }
        }
      }

      //this._termUsageList.Add(termName, termUsage);
      this._termUsageList.Add(_termUri, termUsage);
    }

    public TermUsage FindTerm(string qName)
    {
      return null;
    }

    public TermUsage FindTerm(string localName, string nameSpace)
    {
      string name = string.Format("{0}{1}", nameSpace, localName);

      return (TermUsage)this._termUsageList[name];
      //return (TermUsage)this._termUsageList[localName];      
    }

    public TermUsage FindTermByUri(string uri)
    {
      foreach(string termName in this._termUsageList.Keys)
      {
        TermUsage tu = (TermUsage)this._termUsageList[termName];
        if(tu.Uri.Equals(uri))
          return tu;
      }

      return null;
    }

    public IEnumerator GetEnumerator()
    {
      //return new ApplicationProfile.Enumerator();
      return this._termUsageList.GetValueList().GetEnumerator();
    }

    public void Add(ApplicationProfile ap)
    {
      foreach(object termUsage in ap)
      {
        this.AddTermUsage((TermUsage)termUsage);
      }
    }
  }
}
