using System;
using System.Xml;
using System.Net;
using System.IO;
using System.Threading;

/*
REPOSITORY
Un harvester pu voler ottenere una copia completa di un repository
comprendendo la struttura dei set e tutti i formati di metadati. Per fare questo
bisogna seguire una strategia come la seguente:
* Inviare una richiesta Identify per ricavare la granularit di datestamp pi
fine supportata dal repository.
* Inviare una richiesta ListMetadataFormats per ottenere una lista di tutti i
formati (da assegnare a metadataPrefix) supportati.
* Raccogliere usando le richieste ListRecords per ogni formato
(metadataPrefix) supportato. Essere a conoscenza del fatto che il
repository supporti la granularit di tipo
YYYY-MM-DDThh:mm:ssZ assicura minor sovrapposizione di record
dovuta a cambiamenti durante le raccolte.
* Utilizzando gli elementi setSpec contenuti negli header dei vari record
raccolti,  possibile dedurre la struttura dei set implementati dal
repository.
* Si possono ricostruire i vari item del repository verificando tutti i record
aventi lo stesso valore per il campo unique identifier e considerando tale
valore come identificatore dellitem stesso. Ai vari item raccolti bisogna
assegnare come valori di datestamp, data e ora locali al sistema di raccolta.
* Le informazioni di provenienza, i diritti di utilizzo, ecc., contenute nel
campo <about> possono essere ricostruite al livello di item se tutti i record
appartenenti ad un item specifico hanno il medesimo contenuto per tale
campo. Nel caso in cui tale informazione  differente per ogni record
appartenente allo stesso item, pu essere necessario immagazzinare tale
informazione separatamente per ogni formato di metadato. 
*/

namespace pico.oaipmh2
{
  /// <summary>
  /// Descrizione di riepilogo per oaipmh2.
  /// </summary>
  public class Harvester
  {    
    public int HttpRequestTimeout = 30000; // millisecondi
    public int ListIdentifiersDelay = 50;  // millisecondi
    public string UserAgent = "picovalidator/1.0";
    public string MailFrom = "davide@merlitti.it";

    private string _baseUri = null;
    private int _listedSetCount = 0;

    private const string ParamVerb = "verb";
    private const string ParamIdentifier = "identifier";
    private const string ParamSet = "set";
    private const string ParamMetadataPrefix = "metadataPrefix";
    private const string ParamResumptionToken = "resumptionToken";

    private const string AttributeCursor = "cursor";
    private const string AttributeCompleteListSize = "completeListSize";

    private const string AttributeCode = "code";

    private const string VerbListIdentifiers = "ListIdentifiers";
    private const string VerbGetRecord = "GetRecord";
    private const string VerbListSets = "ListSets";

    private const string NamespaceOAI2 = "http://www.openarchives.org/OAI/2.0/";
    private const string PrefixXsi = "xsi";
    private const string NamespaceXsi = "http://www.w3.org/2001/XMLSchema-instance";

    private const string OaipmhErrorResumptionTokenNotFound = "resumptionToken non trovato";

    public Harvester()
    {
      //
      // TODO: aggiungere qui la logica del costruttore
      //
    }

    public delegate void SetHandler(string setSpec, XmlNode info);

    public delegate void IdentifierHandler(string setSpec, string identifier);

    public delegate void RecordHandler(string setSpec, string identifier, XmlDocument record);

    public delegate void ErrorHandler(string request, string errorInfo);

    public delegate void HttpErrorHandler(string request, string errorInfo);

    private static IdentifierHandler _identifierHandler = null;
    
    private static RecordHandler _recordHandler = null;
    
    private static SetHandler _setHandler = null;

    private static ErrorHandler _errorHandler = null;

    private static HttpErrorHandler _httpErrorHandler = null;

    private static HttpStatusCode _httpStatusCode;
    
    private static string _accept = "text/xml";

    private void Error(string reqUri, string msg)
    {
      if(_errorHandler != null)
        _errorHandler(reqUri, msg);
    }

    private HttpWebRequest CreateRequest(Uri uri)
    {
      HttpWebRequest req = (HttpWebRequest)WebRequest.Create(uri);
      req.UserAgent = UserAgent;
      req.Headers.Add("From", MailFrom);
      req.Accept = _accept;
      req.Timeout = this.HttpRequestTimeout;

      return req;
    }

    private bool CheckHttpResponse(HttpWebResponse res, string content)
    {
      string resUri = res.ResponseUri.AbsoluteUri;

      if(res.StatusCode != HttpStatusCode.OK)
      {
        return false;
      }

      if(res.ContentLength <= 0)
      {
        Error(resUri, "Il contenuto della risposta  vuoto");
        return false;
      }

      if(!res.ContentType.Equals("text/xml"))
      {
        Error(resUri, "Il tipo di contenuto non  'text/xml'");
        return false;
      }

      string charSet = res.CharacterSet;
      string contEnc = res.ContentEncoding;

      return true;
    }

    private bool CheckContent(HttpWebResponse res, string text)
    {
      try
      {
        NameTable nameTable = new NameTable();
        XmlNamespaceManager xmlNamespaceManager = new XmlNamespaceManager(nameTable);
        XmlParserContext xmlParserContext = new XmlParserContext(nameTable, xmlNamespaceManager, null, XmlSpace.Default);

        XmlTextReader r = new XmlTextReader(text, System.Xml.XmlNodeType.Document, xmlParserContext);
        while(r.Read());
        r.Close();
      }
      catch(XmlException xe)
      {
        string errMsg = string.Format("{0}: riga: {1} pos: {2}", xe.Message, xe.LineNumber, xe.LinePosition);
        Error(res.ResponseUri.AbsoluteUri, errMsg);
      }

      return true;
    }

    private bool CheckOaiPmhErrors(string reqUri, ref XmlDocument doc)
    {
      XmlNodeList errorList = doc.SelectNodes("/*[local-name()='OAI-PMH']/*[local-name()='error']");

      if(errorList != null)
      {
        foreach(XmlNode nodeErr in errorList)
        {
          string errorText = nodeErr.InnerText;
          XmlAttribute attrErrCode = nodeErr.Attributes["code"];
          string errorCode = "?";
          if(attrErrCode != null)
          {
            errorCode = attrErrCode.Value;
          }
          string errMsg = string.Format("Error code: {0} ({1})", errorCode, errorText);
          if(_errorHandler != null)
            _errorHandler(reqUri, errMsg);
        }
        return (errorList.Count > 0);
      }

      return false;
    }

    public int GetListedSetsCount()
    {
      return this._listedSetCount;
    }

    public void SetBaseUri(string uri)
    {
      this._baseUri = uri;
    }

    public void SetIdentifierHandler(IdentifierHandler identifierHandler)
    {
      _identifierHandler = identifierHandler;
    }

    public void SetRecordHandler(RecordHandler recordHandler)
    {
      _recordHandler = recordHandler;
    }

    public void SetSetHandler(SetHandler setHandler)
    {
      _setHandler = setHandler;
    }

    public void SetHttpErrorHandler(HttpErrorHandler errorHandler)
    {
      _httpErrorHandler = errorHandler;
    }

    public void SetErrorHandler(ErrorHandler errorHandler)
    {
      _errorHandler = errorHandler;
    }    

    private bool CheckHttpStatus(HttpWebRequest req, HttpWebResponse res)
    {
      if(res.StatusCode != HttpStatusCode.OK)
      {
        if(_httpErrorHandler != null)
          _httpErrorHandler(req.RequestUri.AbsoluteUri, string.Format("Codice {0}: {1}", res.StatusCode,
            res.StatusDescription));

        return true;
      }

      return false;
    }

    private string ReadContent(HttpWebResponse res)
    {
      int bytesRead = 0;
      char[] buffer = new char[512];

      StreamReader reader = new StreamReader(res.GetResponseStream(), System.Text.Encoding.UTF8);
      StringWriter writer = new StringWriter();

      bytesRead = reader.Read(buffer, 0, 512);
      while (bytesRead != 0 ) 
      {
        writer.Write(buffer, 0, bytesRead);
        bytesRead = reader.Read(buffer, 0, 512);
      }

      return writer.ToString();
    }

    private void DumpError(string content)
    {
      FileInfo fi = new FileInfo("errdump.txt");
      StreamWriter errWriter = fi.CreateText();
      errWriter.Write(content);
      errWriter.Close();
    }

    // esegue diverse richieste di ListSets gestendo il resumptionToken
    public void ListSets(string resumptionToken)
    {
      int cursor = 0;
      int completeListSize = 0;
      int listedSets = 0;
      bool hasErrors = false;
      
      this._listedSetCount = 0;

      do
      {
        HttpWebRequest req = CreateRequest(ListSetsUri(resumptionToken));

        HttpWebResponse res = null;
        string xml = "";

        try
        {
          req.Timeout = 30000;
          res = (HttpWebResponse)req.GetResponse();

          hasErrors = CheckHttpStatus(req, res);

          if(!hasErrors)
            xml = this.ReadContent(res);
        }
        catch(WebException webExc)
        {
          System.Console.Error.WriteLine(webExc.Status.ToString());
          return;
        }

        _httpStatusCode = res.StatusCode;

        if(_httpStatusCode != HttpStatusCode.OK)
          return;
        
        XmlDocument doc = new XmlDocument();
      
        try
        {
          doc.LoadXml(xml);

          hasErrors = CheckOaiPmhErrors(req.RequestUri.AbsoluteUri, ref doc);

          if(!hasErrors)
          {
            XmlNamespaceManager nsmgr = new System.Xml.XmlNamespaceManager(doc.NameTable);
            nsmgr.AddNamespace(String.Empty, NamespaceOAI2);
            nsmgr.AddNamespace(PrefixXsi, NamespaceXsi);

            XmlNode nodeResumptionToken = doc.SelectSingleNode("/*[local-name()='OAI-PMH']/*[local-name()='ListSets']/*[local-name()='resumptionToken']");
            
            if(nodeResumptionToken != null)
            {
              XmlAttribute attr = nodeResumptionToken.Attributes["completeListSize"];
              if(attr != null)
                completeListSize = int.Parse(attr.Value);

              attr = nodeResumptionToken.Attributes["cursor", ""];
              if(attr != null)
                cursor = int.Parse(attr.Value);

              resumptionToken = nodeResumptionToken.InnerText;

              if(resumptionToken != null)
              {
                if(resumptionToken.Trim().Length == 0)
                  resumptionToken = null;
              }
            }
                                   
            XmlNodeList sets = doc.SelectNodes("/*[local-name()='OAI-PMH']/*[local-name()='ListSets']/*[local-name()='set']");

            foreach(XmlNode nodeSet in sets)
            {
              XmlNode node1 = nodeSet.SelectSingleNode("./*[local-name()='setSpec']");
              
              string setSpec = node1.InnerText;
              
              if(node1 != null)
                _setHandler(setSpec, nodeSet);
              else
                System.Console.Error.WriteLine("Error in ListSets");
            }

            listedSets += sets.Count;

            this._listedSetCount = listedSets;
          }
        }
        catch(Exception exc)
        {
          string err = exc.Message;

          DumpError(xml);

          System.Console.Out.WriteLine(err);

          hasErrors = true;
        }
      }while(!hasErrors && resumptionToken != null); //(listedSets < completeListSize));
    }

    // esegue diverse richieste di ListIdentifier gestendo il resumptionToken
    public void ListIdentifiers(string setSpec, string metadataPrefix, string resumptionToken)
    {
      int cursor = 0;
      int completeListSize = -1;
      int listedRecords = 0;
      bool hasErrors = false;

      do
      {
        HttpWebRequest req = CreateRequest(ListIdentifiersUri(setSpec, metadataPrefix, resumptionToken));

        HttpWebResponse res = (HttpWebResponse)req.GetResponse();
        
        string xml = this.ReadContent(res);
        
        XmlDocument doc = new XmlDocument();
      
        try
        {
          doc.LoadXml(xml);

          hasErrors = CheckOaiPmhErrors(req.RequestUri.AbsoluteUri, ref doc);

          if(!hasErrors)
          {
            XmlNamespaceManager nsmgr = new System.Xml.XmlNamespaceManager(doc.NameTable);
            nsmgr.AddNamespace(String.Empty, NamespaceOAI2);
            nsmgr.AddNamespace("xsi", NamespaceXsi);
            
            XmlNode nodeResumptionToken = doc.SelectSingleNode("/*[local-name()='OAI-PMH']/*[local-name()='ListIdentifiers']/*[local-name()='resumptionToken']");
            if(nodeResumptionToken != null)
            {
              XmlAttribute attr = nodeResumptionToken.Attributes["completeListSize"];
              if(attr != null)
                completeListSize = int.Parse(attr.Value);

              attr = nodeResumptionToken.Attributes["cursor", ""];
              if(attr != null)
                cursor = int.Parse(attr.Value);

              resumptionToken = nodeResumptionToken.InnerText;

              bool badResumptionToken = false;
              if(resumptionToken == null)
                badResumptionToken = true;
              //else if (resumptionToken.Length == 0)
              //  badResumptionToken = true;
                
              if(badResumptionToken)
              {
                Error(req.RequestUri.AbsoluteUri, "resumptionToken non trovato");
                return;
              }

              if(resumptionToken.Trim().Length == 0)
                resumptionToken = null;
            }

            XmlNodeList headers = doc.SelectNodes("/*[local-name()='OAI-PMH']/*[local-name()='ListIdentifiers']/*[local-name()='header']");

            foreach(XmlNode header in headers)
            {
              XmlNode node1 = header.SelectSingleNode("./*[local-name()='identifier']");
              //XmlNode node2 = header.SelectSingleNode("./*[local-name()='datestamp']");
              //XmlNode node3 = header.SelectSingleNode("./*[local-name()='setSpec']");

              string identifier = node1.InnerText;
              //string datestampVal = node2.InnerText;
              //string setSpecVal = node3.InnerText;

              _identifierHandler(setSpec, identifier);

              XmlDocument rec = GetRecord(identifier, metadataPrefix);

              if(rec != null)
                _recordHandler(setSpec, identifier, rec);
              else
                System.Console.Error.WriteLine("Error in GetRecord of {0}", identifier);              
            }

            listedRecords += headers.Count;
          }
        }
        catch(Exception exc)
        {
          string err = exc.Message;
          System.Console.Out.WriteLine(err);
          System.Console.Out.WriteLine(req.RequestUri);
          hasErrors = true;
        }

        Thread.Sleep(ListIdentifiersDelay);

        if(completeListSize != -1 && listedRecords >= completeListSize)
          break;

      }while(!hasErrors && resumptionToken != null); //(listedRecords < completeListSize));
    }

    public XmlDocument GetRecord(string identifier, string metadataPrefix)
    {
      try
      {
        Uri uri = GetRecordUri(identifier, metadataPrefix);

        HttpWebRequest req = null;
        string xml = "";

        try
        {
          req = CreateRequest(GetRecordUri(identifier, metadataPrefix));

          HttpWebResponse res = (HttpWebResponse)req.GetResponse();

          res.GetResponseStream();

          _httpStatusCode = res.StatusCode;

          if(_httpStatusCode != HttpStatusCode.OK)
            return null;          

          xml = ReadContent(res);
        }
        catch(WebException webExc)
        {
          string msg = webExc.Message;

          DumpError(xml);
        }

        XmlDocument doc = new XmlDocument();

        doc.LoadXml(xml);

        if(CheckOaiPmhErrors(req.RequestUri.AbsoluteUri, ref doc))
        {
          return null;
        }
        
        return doc;
      }
      catch(Exception exc)
      {
        System.Console.Error.WriteLine(exc.Message);
        return null;
      }
    }

    private Uri ListIdentifiersUri(string setSpec, string metadataPrefix, string resumptionToken)
    {
      string tmp = _baseUri;

      if(resumptionToken != null)
        tmp = string.Format("{0}?{1}={2}&resumptionToken={3}",
          _baseUri, ParamVerb, VerbListIdentifiers, resumptionToken);
      else
      {
        if(setSpec != null)
          tmp = string.Format("{0}?{1}={2}&set={3}&metadataPrefix={4}",
            _baseUri, ParamVerb, VerbListIdentifiers, setSpec, metadataPrefix);
        else
          tmp = string.Format("{0}?{1}={2}&metadataPrefix={3}",
            _baseUri, ParamVerb, VerbListIdentifiers, metadataPrefix);
      }
            
      return new Uri(tmp, true);
    }

    private Uri GetRecordUri(string identifier, string metadataPrefix)
    {
      string ret = _baseUri;

      ret = string.Format("{0}?{1}={2}&identifier={3}&metadataPrefix={4}",
        _baseUri, ParamVerb, VerbGetRecord, identifier, metadataPrefix);

      return new Uri(ret, true);
    }

    private Uri ListSetsUri(string resumptionToken)
    {
      string tmp = _baseUri;

      if(resumptionToken != null)
        tmp = string.Format("{0}?{1}={2}&resumptionToken={3}",
          _baseUri, ParamVerb, VerbListSets, resumptionToken);
      else
        tmp = string.Format("{0}?{1}={2}",
          _baseUri, ParamVerb, VerbListSets);

      return new Uri(tmp, true);
    }
  }
}
