using System;
using System.Xml;
using System.Collections;
using System.Text.RegularExpressions;

namespace pico.ap
{
  /// <summary>
  /// Descrizione di riepilogo per Validator.
  /// </summary>
  public class Validator
  {
    ApplicationProfile _ap = null;
    ErrorHandler _errorHandler = null;
    MetadataHandler _metadataHandler = null;
    ArrayList _langs = null;
    Hashtable _dcsv = null;
    Hashtable _pattern = null;
    char ErrorLevel = 'w';
    Hashtable _nameSpace = null;

    public class DcsvComponent
    {
      public string _label = null;
      public string _value = null;
     
      public DcsvComponent(string v)
      {
        this._label = null;
        this._value = v;
      }

      public DcsvComponent(string l, string v)
      {
        this._label = l;
        this._value = v;
      }      
    }

    /*
    ------------------------------------------------------------------------
    Date:         Sat, 26 Feb 2005 09:19:17 +0000
    From: Andy Powell <a.powell@UKOLN.AC.UK>
    Subject: Re: Vocabulary Encoding Scheme vs Syntax Encoding Scheme
    To: DC-USAGE@JISCMAIL.AC.UK
    ------------------------------------------------------------------------

    The rule I'm using is...

    If it is possible to enumerate a finite list of all possible
    valid values then the encoding scheme is a 'vocabulary'.

    Box             Syntax
    DCMIType        Vocabulary
    DDC             Vocabulary
    IMT             Vocabulary
    ISO3166         Vocabulary
    ISO639-2        Vocabulary
    LCC             Vocabulary
    LCSH            Vocabulary
    MESH            Vocabulary
    Period          Syntax
    Point           Syntax
    RFC1766         Vocabulary [1]
    RFC3066         Vocabulary
    TGN             Vocabulary
    UDC             Vocabulary
    URI             Syntax
    W3CDTF          Syntax
    
    Things like RFC1766 [1] look a bit like 'syntaxes' because
    they are made up of component parts - but the fact that you can
    enumerate a list of all possible valid values means (I think)
    that they are 'vocabularies'.

    [1] http://dublincore.org/documents/dcmi-terms/#RFC1766
    */

    public Validator(ApplicationProfile ap)
    {
      _ap = ap;

      // Questi valori andrebbero caricati da un file RDF
      _langs = new ArrayList();
      _langs.Add("it");   // ISO 639-1
      _langs.Add("ita");  // ISO 639-2
      _langs.Add("en");   // ISO 639-1
      _langs.Add("eng");  // ISO 639-2
      _langs.Add("fr");   // ISO 639-1
      _langs.Add("fre");  // ISO 639-2
      _langs.Add("fra");  // ISO 639-2
      _langs.Add("de");   // ISO 639-1
      _langs.Add("deu");  // ISO 639-2
      _langs.Add("ger");  // ISO 639-2
      _langs.Add("es");   // ISO 639-1
      _langs.Add("spa");  // ISO 639-2

      // Questi valori andrebbero caricati da un file RDF
      _dcsv = new Hashtable();      
      _dcsv.Add("NULL", "name?,value?"); // Encoding fittizio per valori non specificati da xsi:type
      //_dcsv.Add("pico:Thesaurus", "!id,!name");
      _dcsv.Add("pico:PostalAddress", "name?,placetype?,placename?,placenumber?,zipcode?,cap?,city,province?,region?,country?");
      _dcsv.Add("pico:ISTAT", "name?,code,year?");
      _dcsv.Add("pico:AAT", "name,value");
      _dcsv.Add("pico:Anchor", "title,url");
      _dcsv.Add("pico:ICONCLASS", "name,value");      
      _dcsv.Add("pico:PRISMResourceType", "name,id");
      _dcsv.Add("pico:PRISMResourceCategory", "name,id");
      _dcsv.Add("pico:ULAN", "name,value");
      _dcsv.Add("pico:UNESCO", "");
      _dcsv.Add("iccd:CD", "tsk,lir?");
      _dcsv.Add("iccd:NCT", "nctr,nctn,ncts?");
      _dcsv.Add("mag:GEN", "stprog,collection?,agency,completeness");
      _dcsv.Add("dcterms:Period", "name?,start,end");
      _dcsv.Add("dcterms:Point", "east,north,elevation?,units?,zunits?,projection,name?");
      _dcsv.Add("dcterms:Box", 
        "northlimit,eastlimit,southlimit,westlimit,uplimit?,downlimit?,units?,zunits?,projection,name?");

      _pattern = new Hashtable();
      _pattern.Add("dcterms:URI", "(([a-zA-Z][0-9a-zA-Z+\\-\\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\\.\\-_!~*'()%]+)?(#[0-9a-zA-Z;/?:@&=+$\\.\\-_!~*'()%]+)?");
      _pattern.Add("dcterms:W3CDTF", "^(\\d{4})(?:(-)?(\\d{2})(?:(?(2)\\-)(\\d{2}))?)?(T\\d{2}:\\d{2}(:\\d{2}(\\.\\d+)?)?(Z|[\\+\\-]\\d{2}:\\d{2}))?$");
      _pattern.Add("pico:Thesaurus", "(([a-zA-Z][0-9a-zA-Z+\\-\\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\\.\\-_!~*'()%]+)?(#[0-9a-zA-Z;/?:@&=+$\\.\\-_!~*'()%]+)?");

      _nameSpace = new Hashtable();
    }

    public void SetNamespace(string prefix, string uri)
    {
      if(this._nameSpace.ContainsKey(prefix))
        this._nameSpace.Remove(prefix);

      this._nameSpace.Add(prefix, uri);      
    }

    public void SetErrorLevel(char errorLevel)
    {
      this.ErrorLevel = errorLevel;
    }

    public void SetErrorHandler(ErrorHandler errorHandler)
    {
      _errorHandler = errorHandler;
    }

    public delegate void ErrorHandler(string errorInfo);

    public void SetMetadataHandler(MetadataHandler metadataHandler)
    {
      _metadataHandler = metadataHandler;
    }

    public delegate void MetadataHandler(XmlQualifiedName qname, XmlQualifiedName xsiType, string xmlLang, string text, ArrayList dcsvComponents);

    /// <summary>
    /// Controlla la validit del nodo xml contenitore rispetto all'ap
    /// </summary>
    /// <param name="nodeRec"></param>
    public void Validate(XmlNode nodeRec)
    {
      foreach(XmlNode node in nodeRec.ChildNodes)
      {
        ArrayList elements = new ArrayList();

        if(node.NodeType == XmlNodeType.Element)
        {
          ValidateElement(node);
          elements.Add(node);
        }
        else
        {
          string errorInfo = string.Format("\"{0}\" {1}", node.Name, "non  un element");

          if(_errorHandler != null)
            _errorHandler(errorInfo);

          Report("E", errorInfo);
        }
      }

      // Controlla gli elementi obbligatori e raccomandati secondo l'ap
      foreach(ApplicationProfile.TermUsage usage in this._ap)
      {
        if(usage.Type == ApplicationProfile.TermUsage.TermUsageType.Element)
        {
          string nsUri = usage.DefinedBy;
          string name = usage.Name;
          string xpq = string.Format("//*[namespace-uri()='{0}'][local-name()='{1}']", 
            nsUri, name);

          if(usage.Obligation == ApplicationProfile.TermUsage.TermUsageObligation.Mandatory)
          {

            XmlNode node = nodeRec.SelectSingleNode(xpq);
            if(node == null)
            {
              string errorInfo = string.Format("Nodo obbligatorio mancante: {0}:{1}",
                nsUri, name);

              if(_errorHandler != null)
                _errorHandler(errorInfo);

              Report("E", errorInfo);
            }
          }
          else if(usage.Obligation == ApplicationProfile.TermUsage.TermUsageObligation.Recommended)
          {
            XmlNode node = nodeRec.SelectSingleNode(xpq);
            if(node == null)
            {
              string errorInfo = string.Format("Nodo raccomandato mancante: {0}:{1}",
                nsUri, name);

              //if(_errorHandler != null)
              //  _errorHandler(errorInfo);

              Report("W", errorInfo);
            }
          }
        }
      }
    }

    private void Report(string msgType, string msgText)
    {
      bool print = true;

      if(this.ErrorLevel == 'e')
        print = (msgType == "E");
        
      if(print)
        System.Console.Out.WriteLine("{0}\t{1}", msgType, msgText);
    }

    /// <summary>
    /// Controlla la validit dell'elemento xml rispetto all'ap.
    /// In particolare controlla se:
    /// 1) il name dell'elemento esiste nell'ap e se il namespace corrisponde a quello dell'ap
    /// 2) l'attributo xml:lang  consentito
    /// 3) il valore di xml:lang  corretto
    /// 4) l'attributo xsi:type  consentito
    /// 5) il valore di xsi:type  corretto per questo elemento
    /// 6) il text rispetta la sintassi prevista per il xsi:type specificato
    /// </summary>
    /// <param name="nodeElem">nodo xml da validare</param>
    private void ValidateElement(XmlNode nodeElem)
    {
      string name = nodeElem.Name;
      string text = nodeElem.InnerText;
      string xmlLang = "";
      string xsiType = "";
      
      XmlAttribute attrXmlLang = nodeElem.Attributes["xml:lang"];
      XmlAttribute attrXsiType = nodeElem.Attributes["xsi:type"];

      if(attrXmlLang != null)
        xmlLang = attrXmlLang.Value;
          
      if(attrXsiType != null)
        xsiType = attrXsiType.Value;

      string elemNamespace = "";
      string elemPrefix = "";
      string elemLocalName = "";

      string typeNamespace = "";
      string typePrefix = "";
      string typeLocalName = "";

      string[] parts = name.Split(':');
      if(parts.Length == 2)
      {
        elemPrefix = parts[0];
        elemLocalName = parts[1];
        elemNamespace = nodeElem.GetNamespaceOfPrefix(elemPrefix);

        // Pezza per gestire il vecchio namespace pico:
        // trasforma l'uri non ufficiale in quella ufficiale
        //if(elemNamespace.StartsWith("http://pico.sns.it/pico/1.0/"))
        //  elemNamespace = string.Format("http://purl.org/pico/1.0/{0}", elemLocalName);
      }

      parts = xsiType.Split(':');
      if(parts.Length == 2)
      {
        typePrefix = parts[0];
        typeLocalName = parts[1];
        typeNamespace = nodeElem.GetNamespaceOfPrefix(typePrefix);

        // Trasforma l'uri non ufficiale in quello ufficiale
        //if(typeNamespace.StartsWith("http://pico.sns.it/pico/1.0/"))
        //  typeNamespace = string.Format("http://purl.org/pico/1.0/{0}", typeLocalName);
      }
      
      ApplicationProfile.TermUsage termUsage = _ap.FindTerm(elemLocalName, elemNamespace);

      if(termUsage != null)
      {
        CheckTermUsage(termUsage, elemLocalName, elemNamespace, xmlLang, xsiType);
      }
      else
      {
        Report("E", string.Format("L'elemento \"{0}\" non  presente nel profilo applicativo", name));        
      }

      ArrayList dcsvComponents = GetDCSV(text);

      //if(dcsvComponents != null)
      //{
        CheckDCSV(nodeElem, xsiType, dcsvComponents);
      //}

      if(xsiType.Length > 0)
      {
        CheckSyntax(nodeElem, xsiType, text);
      }

      if(this._metadataHandler != null)
      {
        XmlQualifiedName qname = new XmlQualifiedName(elemLocalName, elemNamespace);
        XmlQualifiedName qtype = new XmlQualifiedName(typeLocalName, typeNamespace);
        _metadataHandler(qname, qtype, xmlLang, text, dcsvComponents);
      }
    }

    private void CheckSyntax(XmlNode nodeElem, string xsiType, string text)
    {
      if(_pattern.ContainsKey(xsiType))
      {
        string pattern = (string)_pattern[xsiType];

        Regex re = new Regex(pattern);

        if(!re.IsMatch(text))
        {
          _errorHandler(string.Format("Element '{0}', encoding scheme '{1}': il testo non  conforme '{2}'",
            nodeElem.Name, xsiType, text));
        }
      }
    }

    private ArrayList GetDCSV(string text)
    {
      string semicolonPlaceHolder = "@@S@@";
      string equalPlaceHolder = "@@E@@";

      ArrayList dcsvComponents = new ArrayList();

      string tmp = text.Replace("\\;", semicolonPlaceHolder).Replace("\\=", equalPlaceHolder);
      string[] components = tmp.Split(';');
      foreach(string comp in components)
      {
        int i = comp.IndexOf("=");
        if(i > 0)
        {
          string label = comp.Substring(0, i).Trim().ToLower();                    
          //string val = comp.Substring(i + 1).Trim();
          // il valore non deve essere trimmato
          string val = comp.Substring(i + 1);
          try
          {
            dcsvComponents.Add(new DcsvComponent(label, val));
          }
          catch(Exception)
          {
          }
        }
        else
        {
          dcsvComponents.Add(new DcsvComponent(comp));
        }        
      }

      if(dcsvComponents.Count > 0)
        return dcsvComponents;

      return null;
    }

    private void CheckDCSV(XmlNode nodeElem, string xsiType, ArrayList components)
    {
      if(_dcsv.ContainsKey(xsiType))
      {
        string []labels = ((string)_dcsv[xsiType]).Split(',');
        
        // Itera sulle etichette dei componenti dcsv
        if(components != null)
        {
          // Primo ciclo per controllare che non ci siano etichette impreviste e/o ripetute
          foreach(DcsvComponent comp in components)
          {
            string label = comp._label;
            if(label != null)
            {
              int found = 0;
              foreach(string _label in labels)
              {                
                if(_label.TrimEnd('?').Equals(label.ToLower()))
                  found++;
              }
            
              if(found == 0)
                _errorHandler(string.Format("Element '{0}', encoding scheme '{1}': etichetta dcsv non prevista '{2}'",
                  nodeElem.Name, xsiType, label));
              else if(found > 1)
                _errorHandler(string.Format("Element '{0}', encoding scheme '{1}': etichetta dcsv ripetuta '{2}'",
                  nodeElem.Name, xsiType, label));
            }
          }
        }

        // Secondo ciclo per controllare che ci siano tutte le etichette obbligatorie
        foreach(string _label in labels)
        {
          if(_label[_label.Length - 1] != '?')
          {
            bool found = false;
            if(components != null)
            {
              foreach(DcsvComponent comp in components)
              {
                if(comp._label != null)
                {
                  if(comp._label.Equals(_label))
                  {
                    found = true;
                    break;
                  }
                }
              }
            }
            if(!found)
              _errorHandler(string.Format("Element '{0}', encoding scheme '{1}': etichetta obbligatoria mancante '{2}'",
                nodeElem.Name, xsiType, _label));
          }
        }
      }
      else if(xsiType.Length == 0)
      {
        return;
      }      
    }

    private string GetNamespace(string prefix)
    {
      if(this._nameSpace.ContainsKey(prefix))
        return (string)this._nameSpace[prefix];

      return null;
    }

    private void CheckTermUsage(ApplicationProfile.TermUsage termUsage, string localName, string nameSpace, string xmlLang, string xsiType)
    {
      // Se ha un xsi:type, controlla che sia negli encodingScheme
      if(xsiType.Length > 0)
      {
        string[] parts = xsiType.Split(':');
        if(parts.Length == 2)
        {
          string xsiTypePrefix = parts[0];
          string xsiTypeName = parts[1];
          string xsiTypeNamespace = this.GetNamespace(xsiTypePrefix);

          ApplicationProfile.TermUsage xsiTypeUsage = this._ap.FindTerm(xsiTypeName, xsiTypeNamespace);
          if(xsiTypeUsage != null)
          {
            if(!termUsage.HasEncodingScheme(xsiTypeUsage.Uri))
            {
              // Errore: encoding scheme non previsto
              string msg = "Encoding scheme non previsto";
              this._errorHandler(string.Format("Element \"{0}\" xsi:type=\"{1}\" {2}", 
                localName, xsiType, msg));
            }            
          }
          else
          {
            // Errore: xsiType non esiste
            string msg = "Encoding scheme inesistente";
            this._errorHandler(string.Format("Element \"{0}\" xsi:type=\"{1}\" {2}", 
              localName, xsiType, msg));
          }
        }
        else
        {
          // Errore: prefisso dell'encoding scheme non specificato
          string msg = "Prefisso dell'encoding scheme non specificato";
          this._errorHandler(string.Format("Element \"{0}\" xsi:type=\"{1}\" {2}", 
            localName, xsiType, msg));
        }
      }

      // Se ha un xml:lang, controlla che sia valido
      if(xmlLang.Length > 0)
      {
        if(!_langs.Contains(xmlLang.ToLower()))
        {
          string msg = "Lingua non riconosciuta";
          this._errorHandler(string.Format("Element \"{0}\" xml:lang=\"{1}\" {2}", 
            localName, xmlLang, msg));
        }
      }

    }
  }
}